from pathlib import Path
from subprocess import DEVNULL, run
from threading import Thread
from typing import Optional

from loguru import logger

from ._ffmpeg import FFMPEG

TIMESTAMP_FORMAT = "%Y%m%d_%H%M%S"


class Archiver(Thread):
    name = "archiver"
    url: str
    path: Path
    segment_size: int
    segment_filename: str
    segment_format: str
    segment_format_options: Optional[str]
    copy_stream: bool

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        url: str,
        path: Path,
        segment_size: int,
        segment_filename: str,
        segment_format: str,
        segment_format_options: Optional[str],
        copy_stream: bool,
    ):
        Thread.__init__(self)
        self.url = url

        if not path.is_dir():
            raise ValueError(f"archive path '{path}' is not a directory!")
        self.path = path

        self.segment_size = segment_size
        self.segment_filename = segment_filename
        self.segment_format = segment_format
        self.segment_format_options = segment_format_options
        self.copy_stream = copy_stream

    def run(self):
        logger.info("starting archiver")

        cmd = [FFMPEG, "-hide_banner", "-nostats"]

        cmd += ("-re", "-i", self.url)
        cmd += ("-vn",)  # Drop video
        if self.copy_stream:
            logger.info("enabling stream copy")
            cmd += ("-c", "copy")

        cmd += ("-map", "0")
        cmd += ("-map_metadata", "-1")

        cmd += ("-f", "segment")
        cmd += ("-strftime", "1")
        cmd += ("-segment_time", str(self.segment_size))
        cmd += ("-segment_format", self.segment_format)
        if self.segment_format_options:
            cmd += ("-segment_format_options", self.segment_format_options)
        cmd += ("-reset_timestamps", "1")
        cmd += (self.path / f"{self.segment_filename}.{self.segment_format}",)

        run(cmd, check=True, stderr=DEVNULL)
        logger.info("archiver stopped")



import re
from math import isclose
from queue import Queue
from subprocess import DEVNULL, PIPE, Popen
from threading import Thread
from typing import Optional

from loguru import logger

from ._ffmpeg import FFMPEG
from .event import SilenceEvent

NOISE = 0.001

SILENCE_DETECT_RE = re.compile(
    r"\[silencedetect.*\] silence_(start|end): (\d+(?:\.\d+)?)(?: \| silence_duration: (\d+\.\d+))?"
)


def parse_silence_detect(line: str) -> Optional[SilenceEvent]:
    match = SILENCE_DETECT_RE.search(line)
    if match is None:
        return None

    return SilenceEvent(
        kind=match.group(1),
        seconds=float(match.group(2)),
        duration=float(match.group(3)) if match.group(3) else None,
    )


def validate_silence_duration(
    start: SilenceEvent,
    end: Optional[SilenceEvent],
) -> Optional[bool]:
    if end is None:
        return None

    if end.duration is None or end.seconds is None or start.seconds is None:
        return False

    duration_ffmpeg = end.duration
    duration_ours = end.seconds - start.seconds
    is_valid = isclose(duration_ffmpeg, duration_ours, abs_tol=0.01)
    if not is_valid:
        logger.error(
            f"computed duration '{duration_ours}' "
            f"differs from ffmpeg duration '{duration_ffmpeg}'"
        )
    return is_valid


def silence_listener(event_queue: Queue, url: str):
    with Popen(
        (
            *(FFMPEG, "-hide_banner", "-nostats"),
            *("-re", "-i", url),
            "-vn",  # Drop video
            *("-af", f"silencedetect=noise={NOISE}"),
            *("-f", "null", "/dev/null"),
        ),
        bufsize=1,
        stdout=DEVNULL,
        stderr=PIPE,
        text=True,
    ) as process:
        previous = None

        logger.debug("starting to parse stdout")
        for line in process.stderr:  # type: ignore
            event = parse_silence_detect(line)
            if event is None:
                continue

            if event.kind == "end" and previous is not None:
                validate_silence_duration(previous, event)

            logger.info(f"silence {event.kind}: {event.when}")
            event_queue.put(event)
            previous = event


class SilenceListener(Thread):
    name = "silence_listener"
    event_queue: Queue
    url: str

    def __init__(self, event_queue, url):
        Thread.__init__(self)
        self.event_queue = event_queue
        self.url = url

    def run(self):
        logger.info("starting silence listener")
        silence_listener(self.event_queue, self.url)
        logger.info("silence listener stopped")



from queue import Queue
from threading import Event as ThreadEvent
from time import sleep

import httpx
from loguru import logger

from .event import SilenceEvent, StatusEvent


def check_stream(
    event_queue: Queue,
    stop: ThreadEvent,
    url: str,
):
    """
    Check the stream until it is available.
    """
    while not stop.is_set():
        try:
            with httpx.stream("GET", url) as response:
                response.raise_for_status()
                event_queue.put(StatusEvent(kind="up"))
                event_queue.put(SilenceEvent(kind="end", seconds=0.0))
                return

        except (httpx.ConnectError, httpx.HTTPStatusError) as error:
            logger.error(f"could not stream from '{url}'")
            logger.debug(error)
            event_queue.put(StatusEvent(kind="down"))
            sleep(5)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['earhorn']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'httpx>=0.21.1,<0.23.0',
 'loguru>=0.5.3,<0.7.0',
 'lxml>=4.6.3,<4.9.0',
 'more-itertools>=8.10.0,<9.0.0',
 'prometheus-client>=0.13.1,<0.15.0',
 'pydantic>=1.9.0,<2.0.0']

entry_points = \
{'console_scripts': ['earhorn = earhorn.main:cli']}

setup_kwargs = {
    'name': 'earhorn',
    'version': '0.9.0',
    'description': 'Listen, monitor and archive your streams!',
    'long_description': '# earhorn\n\nListen, monitor and archive your streams!\n\n[![](https://mermaid.ink/svg/pako:eNqNlD9PwzAQxb_KySNqFySWqCoDMLAw0AEhgiorvhIriV1dnAJCfHcuzj8nzsDmvHv383Muzo_IrEKRiNpJh_dafpCstpfr1AB4CVJxcJIc4AWNg1waVSKlAmTdGsgdB4k73q7eYbvdLwsx7Ey2QpdjU4ekQO06JkiADV0TWZ-O7aI-NlTCbpflVme436-DQnMCj35_V0PbqvCkDSq4jUN3ppU34AutfxZiSuy1BF4xTkwoq_9HHtx95vYxDj3ntr1ZjlnRi1GMVNy1ZSCsLD92ru50Yd9ablvEsUcEl06Wiq4QkhbHsYU_DQ90aGTsbZdwtlX3PoMzLOdT6tqhmY9m1CK3pCzXl7l71Lx7DLuGWhSnznCn58Z0eGqM0eZjGu0A84ioOtCW1X7pZb5rCTx8abe4X2gUIJElcHb10rLh6A1L4lRh8GCIRxC4nux88xepHfDU4YZtmTWqv96frE8Y3zFSPDQ2zJXoMxYbUSFVUiv-d_20DangT6jCVCS85Ashm9KlIjW_bG3OigM-KO0sieQkyxo3QjbOHr5NJhJHDQ6m_hfYu37_AFEs0XE)](https://mermaid.live/edit#pako:eNqNlD9PwzAQxb_KySNqFySWqCoDMLAw0AEhgiorvhIriV1dnAJCfHcuzj8nzsDmvHv383Muzo_IrEKRiNpJh_dafpCstpfr1AB4CVJxcJIc4AWNg1waVSKlAmTdGsgdB4k73q7eYbvdLwsx7Ey2QpdjU4ekQO06JkiADV0TWZ-O7aI-NlTCbpflVme436-DQnMCj35_V0PbqvCkDSq4jUN3ppU34AutfxZiSuy1BF4xTkwoq_9HHtx95vYxDj3ntr1ZjlnRi1GMVNy1ZSCsLD92ru50Yd9ablvEsUcEl06Wiq4QkhbHsYU_DQ90aGTsbZdwtlX3PoMzLOdT6tqhmY9m1CK3pCzXl7l71Lx7DLuGWhSnznCn58Z0eGqM0eZjGu0A84ioOtCW1X7pZb5rCTx8abe4X2gUIJElcHb10rLh6A1L4lRh8GCIRxC4nux88xepHfDU4YZtmTWqv96frE8Y3zFSPDQ2zJXoMxYbUSFVUiv-d_20DangT6jCVCS85Ashm9KlIjW_bG3OigM-KO0sieQkyxo3QjbOHr5NJhJHDQ6m_hfYu37_AFEs0XE)\n\n## Install\n\nIf you need to listen or archive an Icecast stream, you will need `ffmpeg`:\n\n```sh\nsudo apt install ffmpeg\n```\n\nInstall earhorn from pip:\n\n```sh\npip install earhorn\n```\n\nYou can start archiving an Icecast stream by providing a stream url and an archive path:\n\n```sh\nearhorn \\\n  --stream-url https://stream.example.org/live.ogg \\\n  --archive-path=/to/my/archive\n```\n\nYou can also start exporting the Icecast stats as prometheus metrics by providing an Icecast stats url:\n\n```sh\nearhorn \\\n  --stats-url https://stream.example.org/admin/stats.xml \\\n  --stats-user admin \\\n  --stats-password hackme\n```\n\n### Docker\n\n```sh\ndocker pull ghcr.io/jooola/earhorn\n```\n\n## Usage\n\n```\nUsage: earhorn [OPTIONS]\n\n  See the ffmpeg documentation for details about the `--archive-segment-*` options:\n  https://ffmpeg.org/ffmpeg-formats.html#segment_002c-stream_005fsegment_002c-ssegment\n\nOptions:\n  --listen-port INTEGER           Listen port for the prometheus metrics endpoint.  [default: 9950]\n  --hook PATH                     Path to a custom script executed to handle stream state `events`.\n  --stream-url TEXT               URL to the icecast stream.\n  --stats-url TEXT                URL to the icecast admin xml stats page.\n  --stats-user TEXT               Username for the icecast admin xml stats page.  [default: admin]\n  --stats-password TEXT           Password for the icecast admin xml stats page.\n  --archive-path PATH             Path to the archive storage directory. If defined, the archiver will save the\n                                  `stream` in segments in the storage path.\n  --archive-segment-size INTEGER  Archive segment size in seconds.  [default: 3600]\n  --archive-segment-filename TEXT\n                                  Archive segment filename (without extension).  [default: archive-%Y%m%d_%H%M%S]\n  --archive-segment-format TEXT   Archive segment format.  [default: ogg]\n  --archive-segment-format-options TEXT\n                                  Archive segment format options.\n  --archive-copy-stream           Copy the `stream` without transcoding (reduce CPU usage). WARNING: The stream has to\n                                  be in the same format as the `--archive-segment-format`.\n  --help                          Show this message and exit.\n\n```\n\n## Developmement\n\nTo develop this project, start by reading the `Makefile` to have a basic understanding of the possible tasks.\n\nInstall the project and the dependencies in a virtual environment:\n\n```sh\nmake install\nsource .venv/bin/activate\nearhorn --help\n```\n\n## Releases\n\nTo release a new version, first bump the version number in `pyproject.toml` by hand or by using:\n\n```sh\n# poetry version --help\npoetry version <patch|minor|major>\n```\n\nRun the release target:\n\n```sh\nmake release\n```\n\nFinally, push the release commit and tag to publish them to Pypi:\n\n```sh\ngit push --follow-tags\n```\n',
    'author': 'Joola',
    'author_email': 'jooola@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from prometheus_client import Enum, Gauge, Info, Summary

# Internal

stats_extraction = Summary(
    "earhorn_stats_extraction",
    "Time spent extracting Icecast stats",
)

# Stream

stream_status = Enum(
    "earhorn_stream_status",
    "Whether the stream is up",
    states=["up", "down"],
)
stream_silence = Enum(
    "earhorn_stream_silence",
    "Whether the stream is silent",
    states=["up", "down"],
)

# Icecast server
# https://icecast.org/docs/icecast-latest/server-stats.html

icecast = Info(
    "icecast",
    "Details usually set in the server config, such as: admin, host, "
    "location, server_id, server_start_iso8601.",
)

icecast_client_connections = Gauge(
    "icecast_client_connections",
    "Client connections are basically anything that is not a source "
    "connection. These include listeners (not concurrent, but cumulative), "
    "any admin function accesses, and any static content (file serving) "
    "accesses. This is an accumulating counter.",
)
icecast_clients = Gauge(
    "icecast_clients",
    "Number of currently active client connections.",
)
icecast_connections = Gauge(
    "icecast_connections",
    "The total of all inbound TCP connections since start-up. "
    "This is an accumulating counter.",
)
icecast_file_connections = Gauge(
    "icecast_file_connections",
    "This is an accumulating counter.",
)
icecast_listener_connections = Gauge(
    "icecast_listener_connections",
    "Number of listener connections to mount points. "
    "This is an accumulating counter.",
)
icecast_listeners = Gauge(
    "icecast_listeners", "Number of currently active listener connections."
)
icecast_source_client_connections = Gauge(
    "icecast_source_client_connections",
    "Source client connections are the number of times (cumulative since "
    "start-up, not just currently connected) a source client has connected "
    "to Icecast. This is an accumulating counter.",
)
icecast_source_relay_connections = Gauge(
    "icecast_source_relay_connections",
    "Number of outbound relay connections to (master) icecast servers. "
    "This is an accumulating counter.",
)
icecast_source_total_connections = Gauge(
    "icecast_source_total_connections",
    "Both clients and relays. This is an accumulating counter.",
)
icecast_sources = Gauge(
    "icecast_sources",
    "The total of currently connected sources.",
)
icecast_stats = Gauge(
    "icecast_stats",
    "The total of currently connected STATS clients.",
)
icecast_stats_connections = Gauge(
    "icecast_stats_connections",
    "Number of times a stats client has connected to Icecast. "
    "This is an accumulating counter.",
)

# Icecast sources

icecast_source = Info(
    "icecast_source",
    "Details for the Icecast source, such as: server_name, "
    "server_description, server_type, audio_info, user_agent, "
    "stream_start_iso8601, max_listeners.",
    ["mount"],
)

icecast_source_listener_peak = Gauge(
    "icecast_source_listener_peak",
    "Peak concurrent number of listener connections for this mountpoint.",
    ["mount"],
)
icecast_source_listeners = Gauge(
    "icecast_source_listeners",
    "The number of currently connected listeners.",
    ["mount"],
)
icecast_source_slow_listeners = Gauge(
    "icecast_source_slow_listeners",
    "Number of slow listeners.",
    ["mount"],
)
icecast_source_total_bytes_read = Gauge(
    "icecast_source_total_bytes_read",
    "Total number of bytes received from the source client.",
    ["mount"],
)
icecast_source_total_bytes_sent = Gauge(
    "icecast_source_total_bytes_sent",
    "Total number of bytes sent to all listener connections since "
    "last source connect.",
    ["mount"],
)



from pathlib import Path
from queue import Queue
from signal import SIGINT, SIGTERM, signal
from threading import Event as ThreadEvent
from threading import Thread
from typing import List, Optional

import click
from loguru import logger
from prometheus_client import start_http_server

from .archive import TIMESTAMP_FORMAT, Archiver
from .check import check_stream
from .event import EventHandler, FileHook, PrometheusHook
from .silence import SilenceListener
from .stats import StatsExporter


# pylint: disable=too-many-arguments,too-many-locals
@click.command(context_settings={"max_content_width": 120})
@click.option(
    "--listen-port",
    envvar="LISTEN_PORT",
    help="Listen port for the prometheus metrics endpoint.",
    default=9950,
    show_default=True,
)
@click.option(
    "--hook",
    envvar="HOOK",
    help="Path to a custom script executed to handle stream state `events`.",
    type=click.Path(),
)
@click.option(
    "--stream-url",
    envvar="STREAM_URL",
    help="URL to the icecast stream.",
)
@click.option(
    "--stats-url",
    envvar="STATS_URL",
    help="URL to the icecast admin xml stats page.",
)
@click.option(
    "--stats-user",
    envvar="STATS_USER",
    help="Username for the icecast admin xml stats page.",
    default="admin",
    show_default=True,
)
@click.option(
    "--stats-password",
    envvar="STATS_PASSWORD",
    help="Password for the icecast admin xml stats page.",
)
@click.option(
    "--archive-path",
    envvar="ARCHIVE_PATH",
    help=(
        "Path to the archive storage directory. If defined, the archiver will "
        "save the `stream` in segments in the storage path."
    ),
    type=click.Path(),
)
@click.option(
    "--archive-segment-size",
    envvar="ARCHIVE_SEGMENT_SIZE",
    help="Archive segment size in seconds.",
    default=3600,
    show_default=True,
)
@click.option(
    "--archive-segment-filename",
    envvar="ARCHIVE_SEGMENT_FILENAME",
    help="Archive segment filename (without extension).",
    default=f"archive-{TIMESTAMP_FORMAT}",
    show_default=True,
)
@click.option(
    "--archive-segment-format",
    envvar="ARCHIVE_SEGMENT_FORMAT",
    help="Archive segment format.",
    default="ogg",
    show_default=True,
)
@click.option(
    "--archive-segment-format-options",
    envvar="ARCHIVE_SEGMENT_FORMAT_OPTIONS",
    help="Archive segment format options.",
)
@click.option(
    "--archive-copy-stream",
    envvar="ARCHIVE_COPY_STREAM",
    help=(
        "Copy the `stream` without transcoding (reduce CPU usage). "
        "WARNING: The stream has to be in the same format as the "
        "`--archive-segment-format`."
    ),
    is_flag=True,
)
def cli(
    listen_port: int,
    hook: Optional[str],
    stream_url: Optional[str],
    stats_url: Optional[str],
    stats_user: str,
    stats_password: str,
    archive_path: Optional[str],
    archive_segment_size: int,
    archive_segment_filename: str,
    archive_segment_format: str,
    archive_segment_format_options: Optional[str],
    archive_copy_stream: bool,
):
    """
    \b
    See the ffmpeg documentation for details about the `--archive-segment-*` options:
    https://ffmpeg.org/ffmpeg-formats.html#segment_002c-stream_005fsegment_002c-ssegment
    """

    if stream_url is None and stats_url is None:
        raise click.UsageError("Specify at least one of --stream-url or --stats-url.")

    # Setup stop mechanism
    stop_event = ThreadEvent()

    def stop_handler(_signum, _frame):
        logger.info("stopping...")
        stop_event.set()

    signal(SIGINT, stop_handler)
    signal(SIGTERM, stop_handler)

    # Setup event handler before doing any checks
    event_queue: Queue = Queue()
    event_handler = EventHandler(event_queue, stop_event)

    # Setup prometheus hook
    start_http_server(listen_port)
    event_handler.hooks.append(PrometheusHook())

    # Setup file hook
    if hook is not None:
        event_handler.hooks.append(FileHook(hook))

    event_handler.start()

    # Starting services
    threads: List[Thread] = []

    if stats_url is not None:
        stats_exporter = StatsExporter(
            stop=stop_event,
            url=stats_url,
            auth=(stats_user, stats_password),
        )
        stats_exporter.start()
        threads.append(stats_exporter)

    if stream_url is not None:
        while not stop_event.is_set():
            # Check the stream until it is available
            check_stream(event_queue, stop_event, stream_url)

            silence_listener = SilenceListener(event_queue, stream_url)
            silence_listener.start()
            threads.append(silence_listener)

            if archive_path is not None:
                archiver = Archiver(
                    stream_url,
                    Path(archive_path),
                    archive_segment_size,
                    archive_segment_filename,
                    archive_segment_format,
                    archive_segment_format_options,
                    archive_copy_stream,
                )
                archiver.start()
                threads.append(archiver)

            silence_listener.join()
            archiver.join()

    for thread in threads:
        thread.join()



from threading import Event as ThreadEvent
from threading import Thread
from time import sleep
from typing import Dict, Tuple

import httpx
from loguru import logger
from lxml import etree
from prometheus_client import Gauge

from .prometheus import (
    icecast,
    icecast_client_connections,
    icecast_clients,
    icecast_connections,
    icecast_file_connections,
    icecast_listener_connections,
    icecast_listeners,
    icecast_source,
    icecast_source_client_connections,
    icecast_source_listener_peak,
    icecast_source_listeners,
    icecast_source_relay_connections,
    icecast_source_slow_listeners,
    icecast_source_total_bytes_read,
    icecast_source_total_bytes_sent,
    icecast_source_total_connections,
    icecast_sources,
    icecast_stats,
    icecast_stats_connections,
    stats_extraction,
)


def set_gauge_factory(metric: Gauge):
    def func(element: etree._Element):
        if element.text is not None:
            metric.set(int(element.text))

    return func


def set_gauge_labels_factory(metric: Gauge):
    def func(element: etree._Element, labels: Dict[str, str]):
        if element.text is not None:
            metric.labels(**labels).set(int(element.text))

    return func


ICECAST_SOURCE_MAPPING = {
    "listener_peak": set_gauge_labels_factory(icecast_source_listener_peak),
    "listeners": set_gauge_labels_factory(icecast_source_listeners),
    "slow_listeners": set_gauge_labels_factory(icecast_source_slow_listeners),
    "total_bytes_read": set_gauge_labels_factory(icecast_source_total_bytes_read),
    "total_bytes_sent": set_gauge_labels_factory(icecast_source_total_bytes_sent),
}

ICECAST_SOURCE_INFO_LIST = {
    "server_name",
    "server_description",
    "server_type",
    "audio_info",
    "user_agent",
    "stream_start_iso8601",
}


def extract_icecast_source_stats(element: etree._Element):
    infos: Dict[str, str] = {}
    labels = {"mount": element.attrib.get("mount")}

    for child in element.iterchildren():
        if child.tag in ICECAST_SOURCE_INFO_LIST:
            if child.text is not None:
                infos[child.tag] = child.text

        elif child.tag in ICECAST_SOURCE_MAPPING:
            ICECAST_SOURCE_MAPPING[child.tag](child, labels)

    icecast_source.labels(**labels).info(infos)


ICECAST_MAPPING = {
    "client_connections": set_gauge_factory(icecast_client_connections),
    "clients": set_gauge_factory(icecast_clients),
    "connections": set_gauge_factory(icecast_connections),
    "file_connections": set_gauge_factory(icecast_file_connections),
    "listener_connections": set_gauge_factory(icecast_listener_connections),
    "listeners": set_gauge_factory(icecast_listeners),
    "source_client_connections": set_gauge_factory(icecast_source_client_connections),
    "source_relay_connections": set_gauge_factory(icecast_source_relay_connections),
    "source_total_connections": set_gauge_factory(icecast_source_total_connections),
    "sources": set_gauge_factory(icecast_sources),
    "stats": set_gauge_factory(icecast_stats),
    "stats_connections": set_gauge_factory(icecast_stats_connections),
    "source": extract_icecast_source_stats,
}

ICECAST_INFO_LIST = {
    "admin",
    "host",
    "location",
    "server_id",
    "server_start_iso8601",
}


def extract_icecast_stats(element: etree._Element):
    infos: Dict[str, str] = {}
    for child in element.iterchildren():
        if child.tag in ICECAST_INFO_LIST:
            if child.text is not None:
                infos[child.tag] = child.text

        elif child.tag in ICECAST_MAPPING:
            ICECAST_MAPPING[child.tag](child)

    icecast.info(infos)


@stats_extraction.time()
def extract_xml_stats(blob: str):
    element = etree.fromstring(blob)
    extract_icecast_stats(element)


class StatsExporter(Thread):
    name = "stats_exporter"
    stop: ThreadEvent
    url: str
    auth: Tuple[str, str]
    rate: int

    def __init__(
        self,
        stop: ThreadEvent,
        url: str,
        auth: Tuple[str, str],
        rate: int = 15,
    ):
        Thread.__init__(self)
        self.stop = stop
        self.url = url
        self.auth = auth
        self.rate = rate

    def run(self):
        logger.info(f"starting stats exporter with a {self.rate}s refresh rate ")

        while not self.stop.is_set():
            try:
                logger.trace(f"fetching stats from '{self.url}'")
                response = httpx.get(self.url, auth=self.auth)
                response.raise_for_status()
                extract_xml_stats(response.text)

            except (
                httpx.ConnectError,
                httpx.HTTPStatusError,
                httpx.ReadTimeout,
            ) as error:
                logger.error(f"could not get stats from '{self.url}'")
                logger.debug(error)

            if not self.stop.is_set():
                sleep(self.rate)

        logger.info("stats exporter stopped")
